/*********************************************************************
 * Desktop Station one R3 Sketch for Mr Push R3
 *
 * Copyright (C) 2017 Yaasan
 *
 */

// include the library code:
#include <EEPROM.h>
#include <avr/io.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include "Wire.h"
#include "NewSerial.h"
#include "TrackReporterS88_DS.h"


#define MAX_S88DECODER 1
#define RELPYERROR_300 "300 Command error"
#define RELPYERROR_301 "301 Syntax error"
#define RELPYERROR_302 "302 receive timeout"
#define RELPYERROR_303 "303 Unknown error"
#define RELPYERROR_NONE ""

#define LEDSTATE_STOP 0
#define LEDSTATE_RUN 1
#define LEDSTATE_ERR 2

#define PIN_RUNLED	2

#define ADDR_MAIN		0xC003

#define THRESHOLD_EDC 276  // 1024d * (1.2 / 8) * 9V / 5V
#define THRESHOLD_OV  614  // 1024d * (1.2 / 8) * 20V / 5V

NewSerial SerialDS;


/* Decralation classes */
TrackReporterS88_DS reporter(MAX_S88DECODER);

/* LED functions */
byte gLED_State = LEDSTATE_STOP;
byte gLED_Toggle = 0;

/* Parser */
String request;
String function;
word arguments[4];
word numOfArguments;
boolean result;

/* Analog */
byte gDirection = 0;


//Task Schedule
unsigned long gPreviousL5 = 0; //   50ms interval
unsigned long gPreviousL8 = 0; // 1000ms interval
unsigned long gPreviousL9 = 0; // 2000ms interval



void printString(const char *s, int x, int y);
void ReplyPowerPacket(byte inPower);
word stringToWord(String s);

/***********************************************************************
* 
* Boot setup
* 
* 
************************************************************************/

void setup()
{
	
	
	//Pin mode
	pinMode(A1, INPUT);//Edc
	pinMode(9, OUTPUT);// PWM
	pinMode(10, OUTPUT);//PWM
	
	
	//Timer1 PWM Freq
	TCCR1B = TCCR1B & B11111000 | B00000001;    // set timer 1 divisor to     1 for PWM frequency of 31372.55 Hz

	/* LED turn on */
	changeLED();
	
	/* Clear data of the S88 decoder */
	reporter.refresh();
	
	/* Debug message for PC */
	SerialDS.Init();
	SerialDS.println("--------------------------------------");
	SerialDS.println("Desktop Station shield R1 2017/11/12  ");
	SerialDS.println("--------------------------------------");
	SerialDS.println("100 Ready");
	
	//Reset task
	gPreviousL5 = millis();
	gPreviousL8 = millis();
	
	ShutoffPower();
	
	sei();    //ENABLE INTERRUPTION
	
}


boolean parse() 
{

	String request = SerialDS.getRequest();

	char aTemp[48];
	request.toCharArray(aTemp, 30);

	int lpar = request.indexOf('(');
	if (lpar == -1) 
	{
		return false;
	}

	function = String(request.substring(0, lpar));
	function.trim();

	int offset = lpar + 1;
	int comma = request.indexOf(',', offset);
	numOfArguments = 0;
	
	while (comma != -1) 
	{
		String tmp = request.substring(offset, comma);
		tmp.trim();
		arguments[numOfArguments++] = stringToWord(tmp);
		offset = comma + 1;
		comma = request.indexOf(',', offset);
	}

	int rpar = request.indexOf(')', offset);
	while (rpar == -1) 
	{
		return false;
	}

	if (rpar > offset) 
	{
		String tmp = request.substring(offset, rpar);
		tmp.trim();
		arguments[numOfArguments++] = stringToWord(tmp);
	}

	return true;
}

boolean dispatch() 
{
	byte aValue;
	byte aValueHigh;
	boolean aResult;
	char aTempStr[16];
	word aLocProtocol;
	byte aDir = 0;
	long aSpeed = 0;
	
	if ((function == "setLocoDirection") || ((function == "sLD")))
	{
		if( arguments[0] == ADDR_MAIN)
		{
			if( arguments[1] == 2)
			{
				
				if( gDirection == 0)
				{
					ShutoffPower();
				}
				
				//FWD=1,REV=2を FWD=0,REV=1に変換
				gDirection = 1;
			}
			else
			{
				if( gDirection == 1)
				{
					ShutoffPower();
				}
				
				//FWD固定
				gDirection = 0;
			}
		}
		
		return true;
	
	}
	else if ((function == "setLocoFunction") || ((function == "sLF")))
	{
		return true;
	
	}
	else if ((function == "setTurnout") || ((function == "sTO")))
	{
		return true;
	
	}
	else if ((function == "setPower") || ((function == "sPW")))
	{
		
		if( arguments[0] == 1)
		{
			/* 状態変更 */
			gLED_State = LEDSTATE_RUN;
		}
		else
		{
			ShutoffPower();
			/* 状態変更 */
			gLED_State = LEDSTATE_STOP;
		}
		
		
		/* LED状態の変更 */
		changeLED();
		
		return true;
	
	}
	else if ((function == "setLocoSpeed") || ((function == "sLS")))
	{
		
		if( arguments[0] == ADDR_MAIN)
		{
			SetPWMoutput(gDirection, arguments[1]);
		}
		
		
		return true;
	}
	else if ((function == "reset") || ((function == "sRS")))
	{
		SerialDS.println("100 Ready");
		return true;
	}
	/* reset */
	else if ((function == "setPing") || ((function == "sPI")))
	{
		/* Type(2bytes), Version(2bytes), UID(4bytes) */
		SerialDS.println("@DSG,00,03,00,02,00,00,00,00");
		return true;
	}
	else if ((function == "getLocoConfig") || ((function == "gCV")))
	{
		
		return true;
	}
	else if (function == "getS88")
	{
		int aMaxS88Num = MAX_S88DECODER;
		
		if( arguments[0] > 0)
		{
			aMaxS88Num = arguments[0];
		}
		
		//Send a S88 sensor reply 
		SerialDS.print("@S88,");

		reporter.refresh(aMaxS88Num);

		word aFlags = 0;

		for( int j = 0; j < aMaxS88Num; j++)
		{
		  aFlags = (reporter.getByte((j << 1) + 1) << 8) + reporter.getByte(j << 1);
		  
		  SerialDS.printHEX((word)aFlags);
		  SerialDS.print(",");
		}
			
		SerialDS.println("");
		
		return true;
	} /* getS88 */
	else if ((function == "setLocoConfig") || ((function == "sCV")))
	{
		
		return true;
	}
	else
	{
		return false;
	}
}

void loop()
{
	uint8_t aErrNo = 0;
	word aEdc;
	
	while(1)
	{
		aErrNo = 0;
		
		/* Reset watchdog timer. */
		//wdt_reset(); 
		
		
		char *aReplyText = RELPYERROR_NONE;
		
		if( (millis() - gPreviousL8) >= 200)
		{
			
			//LED Control
			taskLED();
			
			/* Sensor check */
			aEdc = analogRead(A1);
			
			/* OV check */
			if( aEdc > THRESHOLD_OV )
			{
				aErrNo = 2;
			}
			/* LV check */
			if( aEdc < THRESHOLD_EDC )
			{
				aErrNo = 1;
				
			}		
			/* エラー処理 */
			if( aErrNo > 0)
			{
				PowerOffByErr(aErrNo);
			}
		
			//Reset task
			gPreviousL8 = millis();
			
		}
		
		
		if( SerialDS.received())
		{
			
	  		if (parse()) {
				if (dispatch()) {
					aReplyText = "200 Ok";
				} else {
					//SerialDS.println(function);
					aReplyText = RELPYERROR_300;
				}
	  			
			} else {
				aReplyText = RELPYERROR_301;
			}
			
			/* Reply to Desktop Station */
			SerialDS.println(aReplyText);
			
			SerialDS.clearReceive();
		}
		else
		{
			/* Nothing to do */
		}
	}
}

void ShutoffPower(void)
{
	analogWrite(9, 255);
	analogWrite(10, 255);
}

void SetPWMoutput(byte inDir, word inPower)
{
	word aPower = inPower >> 2;
	
	if( aPower >= 255)
	{
		aPower = 255;
	}
	
	//SerialDS.printDEC(aPower);
	//SerialDS.println(",");
	
	switch( inDir)
	{
	case 1:
		analogWrite(10, 255 - aPower);
		analogWrite(9, 255);
		break;
		
	case 0:
	default:
		analogWrite(9, 255 - aPower);
		analogWrite(10, 255);
		break;
	}
}


void PowerOffByErr(uint8_t inErrNo)
{
	ShutoffPower();
	
	ReplyPowerPacket(0);
	
	/* 状態変更 */
	gLED_State = LEDSTATE_ERR;
}

void ReplyPowerPacket(byte inPower)
{
	/* Power */
	SerialDS.print("@PWR,");
	SerialDS.printHEX(inPower);
	SerialDS.println(",");
}


void changeLED(void)
{
	
	
	switch( gLED_State)
	{
	case LEDSTATE_STOP:
		digitalWrite(PIN_RUNLED, 0);
		
		break;
		
	case LEDSTATE_RUN:
		digitalWrite(PIN_RUNLED, 1);
		break;
		
	case LEDSTATE_ERR:
		break;
	}
}

void taskLED(void)
{
	
	
	switch( gLED_State)
	{
	case LEDSTATE_STOP:
		break;
		
	case LEDSTATE_RUN:
		break;
		
	case LEDSTATE_ERR:
		digitalWrite(PIN_RUNLED, gLED_Toggle);
		gLED_Toggle = (~gLED_Toggle) & 1;
		
		break;
	}
}

word stringToWord(String s)
 {
  word result = 0;
  
  for (int i = 0; i < s.length(); i++) {
	result = 10 * result + (s.charAt(i) - '0');
  }
  
  return result;
}